use crate::types::Roll;
use std::{collections::HashMap, iter::FromIterator};
use thiserror::Error;

#[derive(Clone, Debug, Error)]
pub enum Error {
    #[error("Tier out of range {0}")]
    TierOutOfRange(u8),
}

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq, Eq, Hash)]
#[serde(rename_all = "lowercase")]
pub enum PoolType {
    Might,
    Speed,
    Intellect,
    Soulcore,
}

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq)]
#[serde(rename_all = "lowercase")]
pub enum CharacterTypeEdge {
    Flexible,
    Static(HashMap<PoolType, u8>),
}

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq)]
pub struct CharacterType {
    name: String,
    starting_pools: HashMap<PoolType, u8>,
    edge: CharacterTypeEdge,
}

impl CharacterType {
    fn pool(&self, type_: &PoolType) -> u8 {
        self.starting_pools.get(type_).unwrap().clone()
    }

    fn edge(&self, type_: &PoolType) -> u8 {
        match self.edge {
            CharacterTypeEdge::Flexible => 0,
            CharacterTypeEdge::Static(ref edge) => edge.get(type_).unwrap_or(&0).clone(),
        }
    }
}

#[derive(Debug, Serialize, Deserialize, PartialEq)]
#[serde(rename_all = "lowercase")]
pub enum Cost {
    Nothing,
    Constant { stat: PoolType, cost: u8 },
    Variable(PoolType),
}

impl Default for Cost {
    fn default() -> Cost {
        Cost::Nothing
    }
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Tier(u8);

impl Tier {
    pub fn new(val: u8) -> Result<Tier, Error> {
        if val < 1 || val > 6 {
            Err(Error::TierOutOfRange(val))
        } else {
            Ok(Tier(val))
        }
    }
}

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq)]
pub struct Pool {
    pub current: u8,
    pub max: u8,
    pub edge: u8,
}

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq)]
pub struct PointAllocation(HashMap<PoolType, u8>);

#[derive(Debug, Serialize, Deserialize, PartialEq)]
pub struct Action {
    name: String,
    #[serde(default)]
    cost: Cost,
    description: String,
}

#[derive(Debug, Serialize, Deserialize, PartialEq)]
#[serde(rename_all = "lowercase")]
pub enum SkillLevel {
    Inability,
    Trained,
    Specialized,
}

#[derive(Debug, Serialize, Deserialize, PartialEq)]
pub struct Skill {
    name: String,
    description: String,
    level: SkillLevel,
}

#[derive(Debug, Serialize, Deserialize, PartialEq)]
pub struct Enabler {
    name: String,
    description: String,
}

#[derive(Debug, Serialize, Deserialize, PartialEq)]
#[serde(rename_all = "lowercase")]
pub enum Ability {
    Skill(Skill),
    Enabler(Enabler),
    Action(Action),
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Descriptor {
    name: String,
    effects: Vec<Ability>,
}

#[derive(Debug, Serialize, Deserialize, PartialEq)]
pub struct Focus {
    name: String,
    tier1: Vec<Ability>,
    tier2: Vec<Ability>,
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "lowercase")]
pub enum Range {
    Immediate,
    Short,
    Long,
    VeryLong,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Attack {
    weapon: String,
    damage: u8,
    range: Range,
    ammo: Option<u8>,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Defense {
    armor: String,
    defense: u8,
    speed_cost: u8,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct CharacterSheet {
    name: String,
    descriptor: Descriptor,
    character_type: CharacterType,
    focus: Focus,
    flex_edge: Option<PoolType>,
    initial_points: PointAllocation,
    /*
    tier: Tier,
    effort: u8,
    cypher_limit: u8,
    skills: Vec<Skill>,
    attacks: Vec<Attack>,
    defenses: Vec<Defense>,
    equipment: Vec<String>,
    cyphers: Vec<Cypher>,
    */
}

impl CharacterSheet {
    pub fn name(&self) -> String {
        self.name.clone()
    }

    pub fn description(&self) -> String {
        let (first, _) = self.descriptor.name.split_at(1);
        let first = first.to_lowercase();
        let article =
            if first == "a" || first == "e" || first == "i" || first == "o" || first == "u" {
                "an"
            } else {
                "a"
            };
        format!(
            "{} {} {} who {}",
            article, self.descriptor.name, self.character_type.name, self.focus.name
        )
    }

    fn pool(&self, pool: &PoolType) -> Pool {
        let max =
            self.character_type.pool(pool) + self.initial_points.0.get(pool).cloned().unwrap_or(0);
        Pool {
            current: max,
            max,
            edge: self.edge(pool),
        }
    }

    fn edge(&self, pool: &PoolType) -> u8 {
        match self.character_type.edge {
            CharacterTypeEdge::Static(ref edges) => edges.get(pool).cloned().unwrap_or(0).clone(),
            CharacterTypeEdge::Flexible => {
                if self.flex_edge == Some(pool.clone()) {
                    1
                } else {
                    0
                }
            }
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
struct CypherTemplate {
    name: String,
    level: Roll,
    form: String,
    description: String,
}

#[derive(Debug, Serialize, Deserialize)]
struct Cypher {
    name: String,
    level: u8,
    form: String,
    description: String,
}

#[cfg(any(test, feature = "test_data"))]
mod test_data {
    use super::*;

    #[allow(dead_code)]
    pub const GLAIVE: &str = "name: Glaive
starting_pools:
  might: 11
  speed: 10
  intellect: 7
edge:
  static:
    might: 1
    speed: 1";
    pub const JACK: &str = "name: Jack
starting_pools:
  might: 10
  speed: 10
  intellect: 10
edge: flexible
  ";

    pub fn jack() -> CharacterType {
        let mut starting_pools: HashMap<PoolType, u8> = HashMap::new();
        starting_pools.insert(PoolType::Might, 10);
        starting_pools.insert(PoolType::Speed, 10);
        starting_pools.insert(PoolType::Intellect, 10);
        CharacterType {
            name: "Jack".to_owned(),
            starting_pools,
            edge: CharacterTypeEdge::Flexible,
        }
    }

    pub const ADAPTABLE: &str = "
name: Adaptable
effects:
  - action:
      name: Versatile
      description: +2 to any pool, changeable every ten hours
  - skill:
      name: Pleasant Social Interactions
      description: Pleasant Social Interactions
      level: trained
  - skill:
      name: Resilient
      description: Overcoming deprivation, sorrow, or pain
      level: trained
";

    pub fn adaptable() -> Descriptor {
        Descriptor {
            name: "Adaptable".to_owned(),
            effects: vec![
                Ability::Action(versatile()),
                Ability::Skill(pleasant_social()),
                Ability::Skill(overcoming_adversity()),
            ],
        }
    }

    pub fn overcoming_adversity() -> Skill {
        Skill {
            name: "Resilient".to_owned(),
            description: "Overcoming the effects of deprivation, sorrow, or pain".to_owned(),
            level: SkillLevel::Trained,
        }
    }

    pub const VERSATILE: &str = "
name: Versatile
description: +2 to any pool, changeable every ten hours
    ";

    pub fn versatile() -> Action {
        Action {
            name: "Versatile".to_owned(),
            cost: Cost::Nothing,
            description: "+2 to any pool, changeable every ten hours".to_owned(),
        }
    }

    pub const ENTERTAINS: &str = "name: Entertains
tier1:
  - name: Levity
    text: Some text which describes what the player gets
    action: enabler
";

    pub const EXPLORES_DARK_PLACES: &str = "name: Explores Dark Places
tier1:
  - name: Superb Explorer
    skills:
      - skill: searching
        level: trained
      - skill: listening
        level: trained
";

    pub const FOCUSES_MIND_OVER_MATTER: &str = "name: Focuses Mind Over Matter
tier1:
  - name: Deflect Attacks
    cost: { type: \"Constant\", stat: \"Intellect\", cost: 1 }
    text: Trained in speed defense for ten minutes
";

    pub const SILVER_TONGUE: &str = "
name: Speaks with a Silver Tongue
equipment:
  - Book of favorite stories
tier1:
  - name: Poetic License
    skills:
      - name: all social interactions
        level: trained
tier2:
  - name: A Smile and a Word
    text: free level of effort on interaction tasks
";

    pub fn silver_tongue() -> Focus {
        Focus {
            name: "Speaks with a Silver Tongue".to_owned(),
            tier1: vec![Ability::Skill(Skill {
                name: "Poetic License".to_owned(),
                description: "all social interactions".to_owned(),
                level: SkillLevel::Trained,
            })],
            tier2: vec![Ability::Enabler(Enabler {
                name: "A Smile and a Word".to_owned(),
                description: "free level of effort on any interaction task".to_owned(),
            })],
        }
    }

    pub const DEFLECT_ATTACKS: &str = "
name: Deflect Attacks
cost: { constant: { stat: \"intellect\", cost: 1 } }
description: Trained in speed defense for ten minutes
";

    pub const PLEASANT_SOCIAL: &str = "
name: Pleasant social interactions
description: Pleasant social interactions
level: trained
";

    /*
     * pub const IMPARTS_WISDOM
     * This focus is interesting because it allows a character to drop a type ability in favor of
     * "Philosophic Confusion".
     */

    pub fn pleasant_social() -> Skill {
        Skill {
            name: "Pleasant social interactions".to_owned(),
            description: "Pleasant social interactions".to_owned(),
            level: SkillLevel::Trained,
        }
    }

    pub fn opal() -> CharacterSheet {
        let mut initial_points: HashMap<PoolType, u8> = HashMap::new();
        initial_points.insert(PoolType::Might, 1);
        initial_points.insert(PoolType::Speed, 2);
        initial_points.insert(PoolType::Intellect, 3);

        CharacterSheet {
            name: "Opal".to_owned(),
            descriptor: adaptable(),
            character_type: jack(),
            focus: silver_tongue(),
            flex_edge: Some(PoolType::Intellect),
            initial_points: PointAllocation(initial_points),
        }
    }
}

#[cfg(test)]
mod test {
    use super::test_data::*;
    use super::*;
    use serde_yaml;

    #[test]
    fn parses_character_type() {
        let glaive: CharacterType = serde_yaml::from_str(GLAIVE).expect("should deserialize");
        assert_eq!(glaive.name, "Glaive");
        assert_eq!(glaive.pool(&PoolType::Might), 11);
        assert_eq!(glaive.edge(&PoolType::Might), 1);
        assert_eq!(glaive.edge(&PoolType::Intellect), 0);
    }

    #[test]
    fn parses_character_type_with_flexible_edge() {
        let jack: CharacterType = serde_yaml::from_str(JACK).expect("should deserialize");
        assert_eq!(jack.name, "Jack");
    }

    #[test]
    fn parses_skill() {
        let skill: Skill = serde_yaml::from_str(PLEASANT_SOCIAL).expect("should deserialize");
        assert_eq!(skill, pleasant_social());
    }

    #[test]
    fn parses_special_ability() {
        let special: Action = serde_yaml::from_str(VERSATILE).expect("should deserialize");
        assert_eq!(special, versatile());
    }

    #[test]
    fn parses_special_ability_with_cost() {
        let deflect = Action {
            name: "Deflect Attacks".to_owned(),
            cost: Cost::Constant {
                stat: PoolType::Intellect,
                cost: 1,
            },
            description: "Trained in speed defense for ten minutes".to_owned(),
        };
        println!("SPECIAL: {}", serde_yaml::to_string(&deflect).unwrap());
        let action: Action = serde_yaml::from_str(DEFLECT_ATTACKS).expect("should deserialize");
        assert_eq!(action, deflect);
    }

    #[test]
    fn parses_adaptable_descriptor() {
        let descriptor: Descriptor = serde_yaml::from_str(ADAPTABLE).expect("should deserialize");
        assert_eq!(descriptor.effects.len(), 3);
        assert_eq!(descriptor.effects[0], Ability::Action(versatile()));
    }

    /*
    #[test]
    fn parses_enabler_focus() {
    let entertains: Focus = serde_yaml::from_str(ENTERTAINS).expect("should deserialize");
    assert_eq!(
    entertains.tier1,
    DescriptorAbility::Enabler(Enabler {
    name: "Levity".to_owned(),
    text: "Some text which describes what the player gets".to_owned(),
    })
    );
    }
    */

    #[test]
    fn opals_character_sheet() {
        let opal = super::test_data::opal();
        assert_eq!(opal.name(), "Opal");
        assert_eq!(
            opal.description(),
            "an Adaptable Jack who Speaks with a Silver Tongue"
        );
        assert_eq!(
            opal.pool(&PoolType::Might),
            Pool {
                current: 11,
                max: 11,
                edge: 0
            }
        );

        assert_eq!(
            opal.pool(&PoolType::Speed),
            Pool {
                current: 12,
                max: 12,
                edge: 0
            }
        );

        assert_eq!(
            opal.pool(&PoolType::Intellect),
            Pool {
                current: 13,
                max: 13,
                edge: 1
            }
        );
    }
}
