use rusqlite::{params, Connection};
use std::{
    ops::{Deref, DerefMut},
    path::PathBuf,
    sync::{Arc, Mutex},
};

pub struct ManagedConnection<'a> {
    pool: &'a Database,
    connection: Option<Connection>,
}

pub struct Database {
    file_path: PathBuf,
    pool: Arc<Mutex<Vec<Connection>>>,
}

impl Database {
    pub fn new(file_path: PathBuf) -> Result<Database, anyhow::Error> {
        let mut connection = Connection::open(file_path.clone())?;

        let tx = connection.transaction()?;
        let version: i32 = tx.pragma_query_value(None, "user_version", |r| r.get(0))?;
        if version == 0 {
            tx.execute_batch(
                "CREATE TABLE users (id STRING PRIMARY KEY, name TEXT);
                CREATE TABLE invitations (token STRING PRIMARY KEY, user_id STRING, FOREIGN KEY(user_id) REFERENCES users(id));
                CREATE TABLE sessions (token STRING PRIMARY KEY, user_id STRING, FOREIGN KEY(user_id) REFERENCES users(id));
                PRAGMA user_version = 1;",
                )?;
        }
        tx.commit()?;

        Ok(Database {
            file_path,
            pool: Arc::new(Mutex::new(vec![connection])),
        })
    }

    pub fn connect<'a>(&'a self) -> Result<ManagedConnection<'a>, anyhow::Error> {
        let mut pool = self.pool.lock().unwrap();
        match pool.pop() {
            Some(connection) => Ok(ManagedConnection {
                pool: &self,
                connection: Some(connection),
            }),
            None => {
                let connection = Connection::open(self.file_path.clone())?;
                Ok(ManagedConnection {
                    pool: &self,
                    connection: Some(connection),
                })
            }
        }
    }

    pub fn release(&self, connection: Connection) {
        let mut pool = self.pool.lock().unwrap();
        pool.push(connection);
    }
}

impl Deref for ManagedConnection<'_> {
    type Target = Connection;

    fn deref(&self) -> &Connection {
        self.connection.as_ref().unwrap()
    }
}

impl DerefMut for ManagedConnection<'_> {
    fn deref_mut(&mut self) -> &mut Connection {
        self.connection.as_mut().unwrap()
    }
}

impl Drop for ManagedConnection<'_> {
    fn drop(&mut self) {
        self.pool.release(self.connection.take().unwrap());
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use tempfile::NamedTempFile;

    #[test]
    fn it_can_create_users() {
        let path = NamedTempFile::new().unwrap().into_temp_path();
        let database = Database::new(path.to_path_buf()).unwrap();
        let mut connection = database.connect().unwrap();
        let tr = connection.transaction().unwrap();
        tr.execute(
            "INSERT INTO users VALUES(?, ?)",
            params!["abcdefg", "mercer"],
        )
        .unwrap();
        tr.commit().unwrap();

        let connection = database.connect().unwrap();
        let id: Option<String> = connection
            .query_row(
                "SELECT id FROM users WHERE name = ?",
                [String::from("mercer")],
                |row| row.get("id"),
            )
            .unwrap();
        assert_eq!(id, Some(String::from("abcdefg")));
    }
}
