use thiserror::Error;

/// This struct covers *fatal* errors for the application. Cross-functional things like full disks,
/// failing disks, database deadlocks, and so forth, which require that the whole application shut
/// down and that the administrator fix a problem.
#[derive(Debug, Error)]
pub enum FatalError {
    #[error("disk is full")]
    DiskFull,
    #[error("io error: {0}")]
    Io(std::io::Error),
}

pub type AppResult<A, E> = Result<Result<A, E>, FatalError>;

pub fn ok<A, E>(val: A) -> AppResult<A, E> {
    Ok(Ok(val))
}

pub fn error<A, E>(err: E) -> AppResult<A, E> {
    Ok(Err(err))
}

pub fn fatal<A, E>(err: FatalError) -> AppResult<A, E> {
    Err(err)
}

#[cfg(test)]
pub fn maybe_fail<E>(possible_errors: Vec<E>) -> AppResult<(), E> {
    ok(())
}
