use serde::{Deserialize, Serialize};
use std::{
    net::{IpAddr, Ipv4Addr, SocketAddr},
    sync::{Arc, RwLock},
};
use warp::Filter;

mod authentication;
use authentication::{AuthenticationDB, AuthenticationError, MemoryAuth, UserId, Username};

mod database;
mod errors;

#[derive(Debug)]
struct AuthenticationRefused;
impl warp::reject::Reject for AuthenticationRefused {}

fn with_authentication(
    auth_ctx: Arc<RwLock<impl AuthenticationDB>>,
) -> impl Filter<Extract = ((Username, UserId),), Error = warp::Rejection> + Clone {
    let auth_ctx = auth_ctx.clone();
    warp::header("authentication").and_then({
        let auth_ctx = auth_ctx.clone();
        move |auth_header: String| {
            let auth_ctx = auth_ctx.clone();
            async move {
                if auth_header.starts_with("Basic ") {
                    let username = auth_header.split(" ").skip(1).collect::<String>();
                    match auth_ctx
                        .read()
                        .unwrap()
                        .get_user_id(Username::from(username.as_str()))
                    {
                        Ok(Ok(userid)) => Ok((Username::from(username.as_str()), userid)),
                        Ok(Err(_)) => Err(warp::reject::custom(AuthenticationRefused)),
                        Err(err) => panic!("{}", err),
                    }
                } else {
                    Err(warp::reject::custom(AuthenticationRefused))
                }
            }
        }
    })
}

#[derive(Serialize)]
struct ErrorResponse {
    error: String,
}

#[derive(Deserialize)]
struct MakeUserParameters {
    username: String,
}

#[derive(Serialize)]
struct MakeUserResponse {
    userid: String,
}

fn make_user(
    auth_ctx: Arc<RwLock<impl AuthenticationDB>>,
) -> impl Filter<Extract = impl warp::Reply, Error = warp::Rejection> + Clone {
    warp::path!("api" / "v1" / "users")
        .and(warp::put())
        .and(warp::body::json())
        .map(move |params: MakeUserParameters| {
            let mut auth_ctx = auth_ctx.write().unwrap();
            match (*auth_ctx).create_user(Username::from(params.username.as_str())) {
                Ok(Ok(userid)) => warp::reply::json(&MakeUserResponse {
                    userid: String::from(userid),
                }),
                Ok(auth_error) => warp::reply::json(&ErrorResponse {
                    error: format!("{:?}", auth_error),
                }),
                Err(err) => panic!("{}", err),
            }
        })
}

fn list_users(
    auth_ctx: Arc<RwLock<impl AuthenticationDB>>,
) -> impl Filter<Extract = impl warp::Reply, Error = warp::Rejection> + Clone {
    warp::path!("api" / "v1" / "users")
        .and(warp::get())
        .map(move || {
            let auth_ctx = auth_ctx.read().unwrap();
            match (*auth_ctx).list_users() {
                Ok(Ok(users)) => warp::reply::json(
                    &users
                        .iter()
                        .map(|u| String::from(u))
                        .collect::<Vec<String>>(),
                ),
                Ok(auth_error) => warp::reply::json(&ErrorResponse {
                    error: format!("{:?}", auth_error),
                }),
                Err(err) => panic!("{}", err),
            }
        })
}

#[tokio::main]
pub async fn main() {
    let auth_ctx: Arc<RwLock<MemoryAuth>> = Arc::new(RwLock::new(Default::default()));

    let echo_unauthenticated = warp::path!("api" / "v1" / "echo" / String).map(|param: String| {
        println!("param: {}", param);
        warp::reply::json(&vec!["unauthenticated", param.as_str()])
    });

    /*
    let authenticate = warp::path!("api" / "v1" / "auth" / String).map(|param: String| {
        println!("param: {}", param);
        warp::reply::json(&param)
    });
    */

    let echo_authenticated = warp::path!("api" / "v1" / "echo" / String)
        .and(with_authentication(auth_ctx.clone()))
        .map(|param: String, (username, userid)| {
            println!("param: {:?}", username);
            println!("param: {:?}", userid);
            println!("param: {}", param);
            warp::reply::json(&vec!["authed", param.as_str()])
        });

    let filter = list_users(auth_ctx.clone())
        .or(make_user(auth_ctx.clone()))
        .or(echo_authenticated)
        .or(echo_unauthenticated);

    let server = warp::serve(filter);
    server
        .run(SocketAddr::new(
            IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1)),
            8001,
        ))
        .await;
}
